
/* Analyseur syntaxique pour Arith */

%{
  open Ast
%}

%token <int> CST
%token <string> IDENT
%token SET, LET, IN, PRINT, COMMA
%token EOF
%token LP RP
%token PLUS MINUS TIMES DIV
%token EQ

/* Dfinitions des priorits et associativits des tokens */

%nonassoc IN
%left PLUS MINUS
%left TIMES DIV
%nonassoc uminus

/* Point d'entre de la grammaire */
%start prog

/* Type des valeurs retournes par l'analyseur syntaxique */
%type <Ast.pprogram> prog

%%

prog:
| l=list(stmt) EOF { l }
;

stmt:
| SET id=IDENT EQ e=expr { PSet (id, e) }
| SET id=IDENT LP f=separated_list(COMMA, IDENT) RP EQ e=expr
                         { PFun(id,f,e) }
| PRINT e=expr           { PPrint e }
;

expr:
| c=CST                        { PCst c }
| id=IDENT                     { PVar id }
| e1=expr op=binop e2=expr     { PBinop (op, e1, e2) }
| MINUS e=expr %prec uminus    { PBinop (Sub, PCst 0, e) }
| LET x=IDENT EQ e1=expr IN e2=expr
                               { PLetin (x, e1, e2) }
| LP e=expr RP                 { e }
| id=IDENT LP el=separated_list(COMMA, expr) RP
                              { PCall (id, el) }
;

%inline binop:
| PLUS  { Add }
| MINUS { Sub }
| TIMES { Mul }
| DIV   { Div }
;


